
/* Name: main.c
 * Project: WinAmp USB LightCtrl via USB
 * Author: frydaysoft (based on CDC example from Osamu Tamura by Recursion Co., Ltd.)
 * Creation Date: 2006-05-12
 * Tabsize: 4
 * Copyright: (c) 2012 frydaysoft 
 * License: Proprietary, free under certain conditions. See Documentation.
 *
 * 2012-04-11   specific handling for LightCtrl
 * 2006-07-08   removed zero-sized receive block
 * 2006-07-08   adapted to higher baud rate by T.Kitazawa
 *
 */

#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include "oddebug.h"
#include "usbdrv.h"

typedef union usbDWord {
    unsigned long	dword;
    uchar   bytes[4];
} usbDWord_t;
#define UART_DEFAULT_BPS     9600
#define HW_CDC_BULK_OUT_SIZE     8
#define HW_CDC_BULK_IN_SIZE      8


#if (1)
  // ***** Atmega8
  // PC0-PC3, PD4-PD7
  #define InitOutpPort()    PORTC &= 0xF0; DDRC |= 0x0F; PORTD &= 0x0F; DDRD |= 0xF0
  #define SetOutp(val)      PORTC &= 0xF0; PORTC |= (val&0x0F); PORTD &= 0x0F; PORTD |= (val&0xF0)
  // PB.1
  #define InitLedPin()      PORTB &= 0xFD; DDRB |= 0x02
  #define SetLed(val)       PORTB &= 0xFD; PORTB |= (val&0x02)
  #define ToggleLed()       PORTB ^= 0x02
  // PB0 Out-CLK for latch 1
  #define InitLatch1Clk()   PORTB &= 0xFE; DDRB |= 0x01
  #define SetLatch1Clk(val) PORTB &= 0xFE; PORTB |= (val&0x01)
  // PB2 Out-CLK for latch 2
  #define InitLatch2Clk()   PORTB &= 0xFB; DDRB |= 0x04
  #define SetLatch2Clk(val) PORTB &= 0xFB; PORTB |= (val&0x04)

#else

  #error "eval board stk500 !!!"
  // ***** Atmega16 test system
  #define InitOutpPort()     PORTB = 0xFF; DDRB   = 0xFF
  #define SetOutp(val)       PORTB  = ~val 
  #define InitLedPin()       //
  #define SetLed(val)        //
  // Out-CLK for latch not used
  #define InitLatch1Clk()    //
  #define SetLatch1Clk(val)  //
  #define InitLatch2Clk()    //
  #define SetLatch2Clk(val)  //
#endif


typedef enum {
    RX_IDLE  = 0,   // wait for start token
    RX_LEN   = 1,   // wait for length byte
    RX_DATA  = 2,   // wait for all data point out by length byte
    RX_STOP  = 3,   // wait for end token
    RX_READY = 4,   // block finish
    RX_ERROR = 5,   // reserved state to catch RX errors
} T_eRxState;

#define RX_MAX_LEN 18     // we have limited RAM so reduce communication protocol
#define RX_CMD_HEADER 2
#define RX_MAX_DATALEN (RX_MAX_LEN-RX_CMD_HEADER)


uchar m_auc_rx_buf[RX_MAX_LEN];    // rx buffer for state machine
uchar m_rx_pos = 0;                // holds number of reveived data
uchar m_rx_exp = 0;                // holds number of expected user data after block sync
T_eRxState e_rx_state = RX_IDLE;   // statemachine control
uchar m_auc_tx_buf[RX_MAX_LEN];    // tx buffer for state machine
uchar *m_ptx_buf;
uchar m_tx_len = 0;                // holds number remaining TX data
uchar m_sendEmptyTxFrame = 0;


#define MK_LOWER    0x20
#define CMD_VER     'V' //'V' >1V 
#define CMD_LIGHT   'L' //'L' >3L?? 

#define START_TOK '>'
#define END_TOK '\n'


#define POS_START 0
#define POS_LEN   1
#define POS_CMD   2
#define POS_DATA  3

#define RES_CMD_OK  '0'
#define RES_CMD_ERR '1'

uchar auc_version[] = { 'L','C',' ','2', '.', '0', '0' };
uchar auc_conv_tbl[16] = {
    '0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'
};

#define GetHexChar(idx) auc_conv_tbl[idx&0xF]
#define ComAddTailToken(buf, idx)  buf[idx++] = END_TOK



enum {
    SEND_ENCAPSULATED_COMMAND = 0,
    GET_ENCAPSULATED_RESPONSE,
    SET_COMM_FEATURE,
    GET_COMM_FEATURE,
    CLEAR_COMM_FEATURE,
    SET_LINE_CODING = 0x20,
    GET_LINE_CODING,
    SET_CONTROL_LINE_STATE,
    SEND_BREAK
};


static PROGMEM char configDescrCDC[] = {   /* USB configuration descriptor */
    9,          /* sizeof(usbDescrConfig): length of descriptor in bytes */
    USBDESCR_CONFIG,    /* descriptor type */
    67,
    0,          /* total length of data returned (including inlined descriptors) */
    2,          /* number of interfaces in this configuration */
    1,          /* index of this configuration */
    0,          /* configuration name string index */
#if USB_CFG_IS_SELF_POWERED
    (1 << 7) | USBATTR_SELFPOWER,       /* attributes */
#else
    (1 << 7),                           /* attributes */
#endif
    USB_CFG_MAX_BUS_POWER/2,            /* max USB current in 2mA units */

    /* interface descriptor follows inline: */
    9,          /* sizeof(usbDescrInterface): length of descriptor in bytes */
    USBDESCR_INTERFACE, /* descriptor type */
    0,          /* index of this interface */
    0,          /* alternate setting for this interface */
    USB_CFG_HAVE_INTRIN_ENDPOINT,   /* endpoints excl 0: number of endpoint descriptors to follow */
    USB_CFG_INTERFACE_CLASS,
    USB_CFG_INTERFACE_SUBCLASS,
    USB_CFG_INTERFACE_PROTOCOL,
    0,          /* string index for interface */

    /* CDC Class-Specific descriptor */
    5,           /* sizeof(usbDescrCDC_HeaderFn): length of descriptor in bytes */
    0x24,        /* descriptor type */
    0,           /* header functional descriptor */
    0x10, 0x01,

    4,           /* sizeof(usbDescrCDC_AcmFn): length of descriptor in bytes */
    0x24,        /* descriptor type */
    2,           /* abstract control management functional descriptor */
    0x02,        /* SET_LINE_CODING,    GET_LINE_CODING, SET_CONTROL_LINE_STATE    */

    5,           /* sizeof(usbDescrCDC_UnionFn): length of descriptor in bytes */
    0x24,        /* descriptor type */
    6,           /* union functional descriptor */
    0,           /* CDC_COMM_INTF_ID */
    1,           /* CDC_DATA_INTF_ID */

    5,           /* sizeof(usbDescrCDC_CallMgtFn): length of descriptor in bytes */
    0x24,        /* descriptor type */
    1,           /* call management functional descriptor */
    3,           /* allow management on data interface, handles call management by itself */
    1,           /* CDC_DATA_INTF_ID */

    /* Endpoint Descriptor */
    7,           /* sizeof(usbDescrEndpoint) */
    USBDESCR_ENDPOINT,  /* descriptor type = endpoint */
    0x80|USB_CFG_EP3_NUMBER,        /* IN endpoint number */
    0x03,        /* attrib: Interrupt endpoint */
    8, 0,        /* maximum packet size */
    USB_CFG_INTR_POLL_INTERVAL,        /* in ms */

    /* Interface Descriptor  */
    9,           /* sizeof(usbDescrInterface): length of descriptor in bytes */
    USBDESCR_INTERFACE,           /* descriptor type */
    1,           /* index of this interface */
    0,           /* alternate setting for this interface */
    2,           /* endpoints excl 0: number of endpoint descriptors to follow */
    0x0A,        /* Data Interface Class Codes */
    0,
    0,           /* Data Interface Class Protocol Codes */
    0,           /* string index for interface */

    /* Endpoint Descriptor */
    7,           /* sizeof(usbDescrEndpoint) */
    USBDESCR_ENDPOINT,  /* descriptor type = endpoint */
    0x01,        /* OUT endpoint number 1 */
    0x02,        /* attrib: Bulk endpoint */
    8, 0,        /* maximum packet size */
    0,           /* in ms */

    /* Endpoint Descriptor */
    7,           /* sizeof(usbDescrEndpoint) */
    USBDESCR_ENDPOINT,  /* descriptor type = endpoint */
    0x81,        /* IN endpoint number 1 */
    0x02,        /* attrib: Bulk endpoint */
    8, 0,        /* maximum packet size */
    0,           /* in ms */
};


uchar usbFunctionDescriptor(usbRequest_t *rq)
{

    if(rq->wValue.bytes[1] == USBDESCR_DEVICE){
        usbMsgPtr = (uchar *)usbDescriptorDevice;
        return usbDescriptorDevice[0];
    }else{  /* must be config descriptor */
        usbMsgPtr = (uchar *)configDescrCDC;
        return sizeof(configDescrCDC);
    }
}


uchar               sendEmptyFrame;
static uchar        intr3Status;    /* used to control interrupt endpoint transmissions */

static uchar        stopbit, parity, databit;
static usbDWord_t   baud;


/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */

uchar usbFunctionSetup(uchar data[8])
{
    usbRequest_t    *rq = (void *)data;

    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS)
    {    /* class request type */

        if( rq->bRequest==GET_LINE_CODING || rq->bRequest==SET_LINE_CODING )
        {
            return 0xff;
        /*    GET_LINE_CODING -> usbFunctionRead()    */
        /*    SET_LINE_CODING -> usbFunctionWrite()    */
        }
        if(rq->bRequest == SET_CONTROL_LINE_STATE){
#if USB_CFG_HAVE_INTRIN_ENDPOINT3
            /* Report serial state (carrier detect). On several Unix platforms,
             * tty devices can only be opened when carrier detect is set.
             */
            if( intr3Status==0 )
                intr3Status = 2;
#endif
        }
#if 1
        /*  Prepare bulk-in endpoint to respond to early termination   */
        if((rq->bmRequestType & USBRQ_DIR_MASK) == USBRQ_DIR_HOST_TO_DEVICE)
            sendEmptyFrame  = 1;
#endif
    }

    return 0;
}


/*---------------------------------------------------------------------------*/
/* usbFunctionRead                                                          */
/*---------------------------------------------------------------------------*/

uchar usbFunctionRead( uchar *data, uchar len )
{
    // ***** return current config to USB
    data[0] = baud.bytes[0];
    data[1] = baud.bytes[1];
    data[2] = baud.bytes[2];
    data[3] = baud.bytes[3];
    data[4] = stopbit;
    data[5] = parity;
    data[6] = databit;

    return 7;
}


/*---------------------------------------------------------------------------*/
/* usbFunctionWrite                                                          */
/*---------------------------------------------------------------------------*/

uchar usbFunctionWrite( uchar *data, uchar len )
{
    // ***** store LINE_CODING
    baud.bytes[0] = data[0];
    baud.bytes[1] = data[1];
    baud.bytes[2] = data[2];
    baud.bytes[3] = data[3];

    stopbit    = data[4];
    parity     = data[5];
    databit    = data[6];
    return 1;
}

/*---------------------------------------------------------------------------*/
/* ucDecodeHexNibble                                                         */
/*---------------------------------------------------------------------------*/

uchar ucDecodeHexNibble(uchar ucDat)
{
    uchar ucRet = ucDat;
    // a=61h, A=41h  '0'=30h
    if ( ucRet & 0x40 )
    {
        ucRet &= 0x07;
        ucRet += 9;
    }
    ucRet &= 0x0F;
    return ucRet;
}

/*---------------------------------------------------------------------------*/
/* ucDecodeHexByte                                                           */
/*---------------------------------------------------------------------------*/

uchar ucDecodeHexByte(uchar *pucBuf)
{
    uchar ucRet=0;

    if (pucBuf)
    { // a=61h, A=41h  '0'=30h
        ucRet = (ucDecodeHexNibble(*pucBuf) << 4) | ucDecodeHexNibble(*(pucBuf+1));
    } 
    return ucRet;
}


/*---------------------------------------------------------------------------*/
/* usbFunctionWriteOut                                                       */
/*---------------------------------------------------------------------------*/
//  usb -> uC
void usbFunctionWriteOut( uchar *data, uchar len )
{
  uchar uc_tmp;

  if (e_rx_state == RX_READY) return;

  for( ; len; len-- ) 
  {

    uc_tmp = *data++;;  
    if ( uc_tmp == START_TOK )
    {   // in sync
        e_rx_state = RX_IDLE;
    }
    // simple RX block statemachine
    // receives a block with syntax eg: ">9C12345678"
    // '>': start token for block synchronisation
    // byte after '>': lenght of following data
    switch ( e_rx_state )
    {
        case RX_IDLE:
            if ( uc_tmp == START_TOK )
            {   // in sync
                m_auc_rx_buf[0] = uc_tmp;
                m_rx_pos = 1;
                m_rx_exp = 0;
                e_rx_state = RX_LEN;
            }
            break;
        case RX_LEN:
            m_auc_rx_buf[1] = uc_tmp;
            if ( uc_tmp > '9' )
            {   //'a'(61h)-'f' or 'A'(41h)-'F' --> a-f
                uc_tmp |= 0x20;    // makeLower()
                uc_tmp -= 0x57;    // 'a'-'f' --> 10..15 
            }
            else
            {   
                uc_tmp -= '0';     // '0'-'9' --> 0-9
            }
            m_rx_exp = (uc_tmp+RX_CMD_HEADER);
            if ( uc_tmp > RX_MAX_DATALEN )
            {   // invalid length, try new sync
                e_rx_state = RX_IDLE;
            }
            else
            {   // length valid, wait for user data
                m_rx_pos = 2;
                e_rx_state = RX_DATA;
            }
            break;
        case RX_DATA:               
            if (m_rx_pos < (RX_MAX_LEN-1))
            {   // copy user data
                m_auc_rx_buf[m_rx_pos] = uc_tmp;
                if ( (++m_rx_pos) >= m_rx_exp )
                {   // block seems to be valid and complete
                    e_rx_state = RX_STOP;
                }
            }
            else
            {   // buffer overflow, try new sync
                e_rx_state = RX_IDLE;
            }            
            break;
        case RX_STOP:
            if (uc_tmp == END_TOK)
            {   // copy user data
                e_rx_state = RX_READY;
                //usbDisableAllRequests();
                return;
            }
            else
            {   // buffer overflow, try new sync
                e_rx_state = RX_IDLE;
            }            
            break;
        default:
            break;
    }
  }
}


/*---------------------------------------------------------------------------*/
/* uartPoll                                                                  */
/*---------------------------------------------------------------------------*/
//  uC -> usb
void uartPoll(void)
{
  uchar uc_tmp;

  if ( e_rx_state == RX_READY )
  {
    if ( !(m_tx_len) )
    {
      ToggleLed();
      m_auc_tx_buf[POS_START] =  m_auc_rx_buf[POS_START];
      m_auc_tx_buf[POS_CMD]   =  m_auc_rx_buf[POS_CMD];

      switch(m_auc_rx_buf[POS_CMD])
      {   // ***** 'V' >1V      return: version data
          case CMD_VER:   
              for (m_tx_len=0; m_tx_len<sizeof(auc_version); m_tx_len++)
              {
                  m_auc_tx_buf[m_tx_len+POS_DATA] = auc_version[m_tx_len];
              }
              m_tx_len += POS_DATA;
              m_auc_tx_buf[POS_LEN] = GetHexChar((m_tx_len-2));
              break;
          // ***** //'L' >3L??   return: >3l:0
          // ***** //'L' >5L???? return: >3l:0
          case CMD_LIGHT: 
              uc_tmp = ucDecodeHexByte(&m_auc_rx_buf[POS_DATA]);
              SetOutp(uc_tmp);
              SetLatch1Clk(1);
              SetLatch1Clk(0);
              if (m_auc_rx_buf[POS_LEN] == '5')
              {
                uc_tmp = ucDecodeHexByte(&m_auc_rx_buf[POS_DATA+2]);
                SetOutp(uc_tmp);
                SetLatch2Clk(1);
                SetLatch2Clk(0);
              }
              m_auc_tx_buf[POS_LEN] = '3';
              m_auc_tx_buf[POS_DATA] = ':';
              m_auc_tx_buf[POS_DATA+1] = RES_CMD_OK; // ok
              m_tx_len = 5;
              break;
          default:
              m_tx_len = 0;
              //usbEnableAllRequests();
              break;
      }
      if ( m_tx_len )
      {   // ***** send response
          m_auc_tx_buf[POS_CMD] += MK_LOWER;
          ComAddTailToken(m_auc_tx_buf, m_tx_len);
      }
      e_rx_state = RX_IDLE;
      //usbEnableAllRequests();
      m_ptx_buf = &m_auc_tx_buf[0];
      //m_tx_len = 0;
    }

  }

  if( usbInterruptIsReady() && (m_tx_len || m_sendEmptyTxFrame) )
  {
    if (m_sendEmptyTxFrame)
    {
      usbSetInterrupt(NULL, 0);
      m_sendEmptyTxFrame = 0;
    }
    else
    {
      if( m_tx_len > HW_CDC_BULK_OUT_SIZE ) 
      {
        usbSetInterrupt(m_ptx_buf, HW_CDC_BULK_OUT_SIZE);
        m_ptx_buf += HW_CDC_BULK_OUT_SIZE;
        m_tx_len -= HW_CDC_BULK_OUT_SIZE;
      }
      else
      {
        // send an empty block after last data block to indicate transfer end
        m_sendEmptyTxFrame = (m_tx_len==HW_CDC_BULK_OUT_SIZE)? 1:0;
        usbSetInterrupt(m_ptx_buf, m_tx_len);
        m_tx_len  = 0;
      }
    }
  }
}

/*---------------------------------------------------------------------------*/
/* UartInit                                                                  */
/*---------------------------------------------------------------------------*/
void UartInit(void)
{
  m_rx_pos = 0;
  m_rx_exp = 0;
  e_rx_state = RX_IDLE;
  m_tx_len = 0;
  m_sendEmptyTxFrame = 0;
}


/*---------------------------------------------------------------------------*/
/* hardwareInit                                                              */
/*---------------------------------------------------------------------------*/

static void hardwareInit(void)
{
    /*    USART configuration    */
    baud.dword  = UART_DEFAULT_BPS;
    stopbit = 0;
    parity  = 0;
    databit = 8;

    InitOutpPort();
    InitLedPin();
    SetOutp(0x00);
    SetLatch1Clk(0);
    SetLatch2Clk(0);
    SetLed(0x00);
    UartInit();

    /* activate pull-ups except on USB lines */
    USB_CFG_IOPORT   = (uchar)~((1<<USB_CFG_DMINUS_BIT)|(1<<USB_CFG_DPLUS_BIT));
    /* all pins input except USB (-> USB reset) */
#ifdef USB_CFG_PULLUP_IOPORT    /* use usbDeviceConnect()/usbDeviceDisconnect() if available */
    USBDDR    = 0;    /* we do RESET by deactivating pullup */
    usbDeviceDisconnect();
#else
    USBDDR    = (1<<USB_CFG_DMINUS_BIT)|(1<<USB_CFG_DPLUS_BIT);
#endif

    /* 250 ms disconnect */
    wdt_reset();
    _delay_ms(250);

#ifdef USB_CFG_PULLUP_IOPORT
    usbDeviceConnect();
#else
    USBDDR    = 0;      /*  remove USB reset condition */
#endif

}


/*---------------------------------------------------------------------------*/
/* main                                                                      */
/*---------------------------------------------------------------------------*/

int main(void)
{
#if 0
    // ***** simple testcode
    unsigned char   val;
    unsigned short  to;

    to = 0;
    val = 0;
    SetOutp(0x00);
    InitOutpPort();  // set outputs

    while(1)
    {
        while(--to) {to = to;}
        SetOutp(val);
        val++;
    }
#endif
    
    wdt_enable(WDTO_1S);
    odDebugInit();
    hardwareInit();
    usbInit();

    intr3Status = 0;
    sendEmptyFrame  = 0;

    sei();
    for(;;)
    {    /* main event loop */
        wdt_reset();
        usbPoll();
        uartPoll();

#if USB_CFG_HAVE_INTRIN_ENDPOINT3
        /* We need to report rx and tx carrier after open attempt */
        if(intr3Status != 0 && usbInterruptIsReady3())
        {
            static uchar serialStateNotification[10] = {0xa1, 0x20, 0, 0, 0, 0, 2, 0, 3, 0};

            if(intr3Status == 2)
            {
                usbSetInterrupt3(serialStateNotification, 8);
            }
            else
            {
                usbSetInterrupt3(serialStateNotification+8, 2);
            }
            intr3Status--;
        }
#endif
    }

    return 0;
}

